#!/usr/bin/env python
# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
import numpy as np
import skimage.filters as filters

from .. misc import gaussian_line, generate_test_image
from .. hough import HoughTransform


def single_segment_fitting_test(
    image_shape=(400,600),
    intensity=1.25, sigma=1.0, width=4.0, band_width=10.0):
  Ny,Nx = image_shape
  x0,y0 = np.random.uniform(20,Nx-20), np.random.uniform(20,Ny-20)
  x1,y1 = np.random.uniform(20,Nx-20), np.random.uniform(20,Ny-20)

  image = intensity*gaussian_line((Ny,Nx), (y0,x0), (y1,x1), width)\
          +sigma*np.random.normal(size=(Ny,Nx))
  fig = plt.figure(figsize=(16,9))
  ax1 = fig.add_subplot(221)
  ax1.imshow(image)
  ax1.scatter(x0,y0, color=(1.0,0.5,0.5))
  ax1.scatter(x1,y1, color=(1.0,0.5,0.5))
  ax1.invert_yaxis()

  ht = HoughTransform(image, sigma)
  theta = np.arctan2(y1-y0,x1-x0)
  cos,sin = np.cos(theta),np.sin(theta)
  r = (sin*(x0-(Nx-1)/2.) - cos*(y0-(Ny-1)/2.))

  el = ht.elements_along(r,theta,band_width)
  res = ht.fit(r,theta,band_width)
  dm1,dm2 = res[2],res[3]

  ax2 = fig.add_subplot(222)
  ax2.imshow(image * ht.mask(r, theta, band_width))
  ax2.invert_yaxis()
  p1 = ht.locate(r, theta, dm1)
  p2 = ht.locate(r, theta, dm2)
  ax2.scatter(p1[0]+(Nx-1)/2,p1[1]+(Ny-1)/2, color=(0.3,0.3,1.0))
  ax2.scatter(p2[0]+(Nx-1)/2,p2[1]+(Ny-1)/2, color=(0.3,0.3,1.0))

  ax3 = fig.add_subplot(223)
  ax3.plot(el[0],el[1])
  ax3.plot(el[0],ht.ramp_sigmoid(el[0],res))

  prof = ht.profile(r, theta, dm1, dm2, 2*band_width)
  ax4 = fig.add_subplot(224)
  ax4.plot(prof[0],prof[1])

  fig.tight_layout()
  plt.show()


if __name__=='__main__':
  from argparse import ArgumentParser as ap
  parser = ap(description='A fitting procedure test with a single segment.')

  parser.add_argument(
    '-S', '--shape', dest='shape', type=int, nargs=2, default=(400,600),
    help='image size')
  parser.add_argument(
    '-i', '--intensity', dest='intensity', type=float, default=1.25,
    help='line intensity')
  parser.add_argument(
    '-s', '--sigma', dest='sigma', type=float, default=1.00,
    help='white noise')
  parser.add_argument(
    '-w', '--width', dest='width', type=float, default=4.0,
    help='line width')
  parser.add_argument(
    '-b', '--band-width', dest='band_width', type=float, default=6.0,
    help='extraction band width')

  args = parser.parse_args()

  single_segment_fitting_test(
    image_shape=args.shape, intensity=args.intensity,
    sigma=args.sigma, width=args.width, band_width=args.band_width)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
import scipy.optimize as opt
import itertools


class HoughTransform(object):
  def __init__(self, image, sigma):
    self.image = image
    self.sigma = sigma
    ny,nx = image.shape
    self.center = (ny-1)/2.0,(nx-1)/2.0
    self.axis1 = np.arange(nx).reshape((1,nx))-self.center[1]
    self.axis2 = np.arange(ny).reshape((ny,1))-self.center[0]
    self.radius = np.sqrt(nx**2+ny**2)/2.0
    self.append_x = self.radius * np.array([-1,1])
    self.append_y = np.zeros(2)

  def score_box(self, rarr, tarr, width):
    nr,nt = rarr.size,tarr.size
    score = []
    for r,t in itertools.product(rarr,tarr):
      score.append(self.snr(r,t,width))
    return np.array(score).reshape((nr,nt))

  def p0(self, r, theta):
    return np.array([r*np.sin(theta),-r*np.cos(theta)])

  def n0(self, r, theta):
    return np.array([np.cos(theta), np.sin(theta)])

  def dl(self, r, theta):
    cos,sin = np.cos(theta),np.sin(theta)
    return (-sin*self.axis1+cos*self.axis2+r)

  def dm(self, r, theta):
    cos,sin = np.cos(theta),np.sin(theta)
    return cos*self.axis1+sin*self.axis2

  def dmlim(self, r, theta):
    ny,nx = self.image.shape
    p = np.tile(self.p0(r,theta), 2)
    b = np.tile(self.n0(r,theta),2)
    b[np.abs(b)<1e-8] = 1e-8
    X = (np.array([-(nx-1)/2,-(ny-1)/2,(nx-1)/2,(ny-1)/2])-p)/b
    Xm,Xp = X[X<0], X[X>0]
    return np.max(Xm),np.min(Xp)

  def locate(self, r, theta, dm):
    return self.p0(r,theta)+dm*self.n0(r,theta)

  def mask(self, r, theta, w):
    return np.abs(self.dl(r,theta))<w

  def elements_along(self, r, theta, w):
    mask = self.mask(r,theta,w)
    dm,px = self.dm(r, theta)[mask], self.image[mask]
    dm = np.append(dm, self.append_x)
    px = np.append(px, self.append_y)
    index = dm.argsort()
    elem = np.vstack((dm[index],np.cumsum(px[index])))
    return elem

  def snr(self, r, theta, w):
    E = self.elements_along(r, theta, w)
    N,m,M = E[1].size,E[1].min(),E[1].max()
    return (M-m)/self.sigma/np.sqrt(N)

  def fit(self, r, theta, w):
    E = self.elements_along(r, theta, w)
    N,m,M = E[1].size,E[1].min(),E[1].max()
    s = self.sigma*np.sqrt(1+np.arange(N))
    dm,dM = self.dmlim(r, theta)
    bounds = opt.Bounds([-np.inf,-np.inf,dm,dm],[np.inf,np.inf,dM,dM])
    init = [m, (M-m)/N, dm, dM-dm]
    func = lambda p: np.sum((E[1]-self.ramp_sigmoid(E[0],p))**2/s**2)
    res = opt.minimize(func, init, method='L-BFGS-B',bounds=bounds)
    return np.hstack((res.x, np.sqrt(res.fun)/N))

  def profile(self, r, theta, dm1, dm2, w):
    dm = self.dm(r,theta)
    mask = (self.mask(r,theta,w)) & (dm>dm1) & (dm<dm2)
    dl = self.dl(r, theta)[mask]
    px = self.image[mask]
    hist,edge = np.histogram(dl, weights=px/(dm2-dm1), bins=int(4*w))
    edge = edge[1:]-np.diff(edge)
    return np.vstack((edge,hist))

  def ramp_sigmoid(self, x, p):
    return p[0]+p[1]*np.clip((x-p[2]), 0, max(p[3]-p[2],1e-8))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
import numpy as np
import skimage.filters as filters

from .. misc import gaussian_line, generate_test_image
from .. hough import HoughTransform


def hough_transformation_score_map_test(
    image_shape, grid, num_line, intensity, sigma, width):
  Ny,Nx = image_shape
  image = generate_test_image(image_shape, num_line, intensity, width)

  fig = plt.figure(figsize=(8,12))
  ax1 = fig.add_subplot(211)
  ax1.imshow(image)
  ax1.invert_yaxis()

  N,M = grid
  ht = HoughTransform(image, sigma)
  R = np.sqrt(Nx**2+Ny**2)/2.0
  r = np.linspace(-R,R,N)
  t = np.linspace(0,np.pi,M)
  box = ht.score_box(r,t,(2*R)/N)
  ax2 = fig.add_subplot(212)
  logbox = np.log(1+np.abs(box))*np.sign(box)
  ax2.imshow(logbox, extent=(0,180,-R,R), aspect='auto')

  plt.show()


if __name__=='__main__':
  from argparse import ArgumentParser as ap
  parser = ap(description='A Hough-Transformation test.')

  parser.add_argument(
    '-S', '--shape', dest='image_shape', type=int, nargs=2, default=(200,300),
    help='image size')
  parser.add_argument(
    '-n', '--grid', dest='grid', type=int, nargs=2, default=(181,181),
    help='number of grids')
  parser.add_argument(
    '-l', '--line', dest='num_line', type=int, default=2,
    help='number of lines')
  parser.add_argument(
    '-i', '--intensity', dest='intensity', type=float, default=5.0,
    help='line intensity')
  parser.add_argument(
    '-s', '--sigma', dest='sigma', type=float, default=1.00,
    help='white noise')
  parser.add_argument(
    '-w', '--width', dest='width', type=float, default=3.0,
    help='line width')

  args = parser.parse_args()

  hough_transformation_score_map_test(
    image_shape=args.image_shape, grid=args.grid,
    num_line=args.num_line, intensity=args.intensity,
    sigma=args.sigma, width=args.width)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np


def gaussian_line(shape,start,end,width):
  ''' Draw a gaussian-blurred line in an image.

  Parameters:
    shape (list): a (ny,nx)-shape of a generated image.
    start (list): the starting point of the line segment.
    end (list): the end point of the line segment.
    width (float): the width of line segments in FWHM.

  Return:
    An image with a line segment.
  '''
  ny,nx = shape
  y0,x0 = start
  y1,x1 = end
  length = np.sqrt((x1-x0)**2+(y1-y0)**2)
  cos,sin = (x1-x0)/length, (y1-y0)/length

  ay = np.arange(ny).reshape((ny,1))
  ax = np.arange(nx).reshape((1,nx))

  dl = np.abs(-sin*(ax-x0)+cos*(ay-y0))
  d0 = cos*(ax-x0)+sin*(ay-y0)
  d1 = cos*(ax-x1)+sin*(ay-y1)
  dx = dl**2*(d1*d0<0) \
    + ((d0<0)&(d1<0))*((ax-x0)**2+(ay-y0)**2) \
    + ((d0>0)&(d1>0))*((ax-x1)**2+(ay-y1)**2)
  sig = 4.*np.log(2.)/width**2

  line = np.exp(-sig*dx**2)
  return line/line.sum()*length


def generate_test_image(shape, n_line=1, intensity=5.0, width=3.0):
  ''' Generate an image with line segments.

  Parameters:
    shape (list): a (ny,nx)-shape of a generated image.
    n_line (int): the number of line segments.
    intensity (float): the line intensity of line segments.
    width (float): the width of line segments in FWHM.

  Return:
    An image with generated line segments.
  '''
  ny,nx = shape
  vy = np.random.uniform(0,ny,(2*n_line))
  vx = np.random.uniform(0,nx,(2*n_line))
  image = np.zeros(shape)
  for n in range(n_line):
    start = vy[2*n],vx[2*n]
    end   = vy[2*n+1],vx[2*n+1]
    image += intensity*gaussian_line(shape, start, end, width)

  image += np.random.normal(size=(ny,nx))
  return image



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from glob import glob
from setuptools import setup, find_packages
from setuptools import Extension
import os,sys,re


with open('README.md', 'r') as fd:
  version = '0.1.0'
  author = 'Ryou Ohsawa'
  email = 'ohsawa@ioa.s.u-tokyo.ac.jp'
  description = 'falsedto: Faint Line Segment Detection Tool'
  long_description = fd.read()
  license = 'MIT'

classifiers = [
  'Development Status :: 3 - Alpha',
  'Environment :: Console',
  'Intended Audience :: Science/Research',
  'License :: OSI Approved :: MIT License',
  'Operating System :: POSIX :: Linux',
  'Programming Language :: Python :: 3',
  'Programming Language :: Python :: 3.7',
  'Topic :: Scientific/Engineering :: Astronomy']


USE_CYTHON = False

def get_filename(name):
  if os.path.exists('{}.pyx'.format(name)):
    USE_CYTHON = True
    return ['{}.pyx'.format(name),]
  else:
    USE_CYTHON = False
    if os.path.exists('{}.cpp'.format(name)):
      filename = ['{}.cpp'.format(name),]
    elif os.path.exists('{}.cxx'.format(name)):
      filename = ['{}.cxx'.format(name),]
    elif os.path.exists('{}.cc'.format(name)):
      filename = ['{}.cc'.format(name),]
    else:
      raise RuntimeError('no file matches name "{}"'.format(name))


if __name__ == '__main__':
  try:
    import numpy
  except ImportError:
    raise SystemExit('NumPy is not available.')


  depends      = None
  libraries    = ['m',]
  include_dirs = [numpy.get_include(),]
  compile_args = ['-std=c++11',]

  ext_modules = [
    Extension(
      'falsedto.hough', language='c++', sources=get_filename('hough'),
      libraries=libraries, include_dirs=include_dirs,
      depends=depends, extra_compile_args=compile_args),
  ]

  if USE_CYTHON:
    from Cython.Build import cythonize
    ext_modules = cythonize(ext_modules, language_level=3)

  setup(
    name='falsedto',
    version=version,
    author=author,
    author_email=email,
    maintainer=author,
    maintainer_email=email,
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://bitbucket.org/ryou_ohsawa/falsedt/src/master/',
    license=license,
    packages=find_packages(),
    classifiers=classifiers,
    install_requires=['numpy','matplotlib','astropy',],
    ext_modules=ext_modules)
